import typing
import collections.abc

from .. import contracts
from .generic import success_shutdown_instance, failure_shutdown_instance
from .provider import DependencyProvider


def _auto_detect_contract(
    instance: contracts.DependencyBindingInstance,
    contract: contracts.DependencyBindingContract | None
) -> contracts.DependencyBindingContract | None:
    if contract:
        return contract
    parent = None
    for cls in instance.__mro__:
        if cls == typing.Protocol:
            return parent
        parent = cls


class BindingChain(contracts.DependencyBindingChain):
    def __init__(self, dependency_registry: contracts.DependencyBingingRegistry):
        self._dependency_registry = dependency_registry

    def filter(self, clause: contracts.DependencyBindingChainFilterClause) -> 'BindingChain':
        self._dependency_registry = {
            contract: record
            for contract, record in self._dependency_registry.items()
            if clause(contract, record)
        }
        return self

    def map(
        self,
        key_clause: contracts.DependencyBindingChainMapClause,
        value_clause: contracts.DependencyBindingChainMapClause
    ) -> 'BindingChain':
        self._dependency_registry = {
            key_clause(contract, record): value_clause(contract, record)
            for contract, record in self._dependency_registry.items()
        }
        return self

    def items(self) -> contracts.DependencyBingingRegistry:
        return self._dependency_registry

    def values(self) -> collections.abc.Iterable[contracts.DependencyBindingRecord]:
        return self._dependency_registry.values()

    def keys(self) -> collections.abc.Iterable[contracts.DependencyBindingContract]:
        return self._dependency_registry.keys()


class DependencyContainer(contracts.DependencyContainer):
    def __init__(self):
        self._dependency_registry = {}
        self._callbacks = {}
        self._instance_registry = {}

    def register_dependency(
        self,
        instance: contracts.DependencyBindingInstance,
        contract: contracts.DependencyBindingContract,
        binding_type: contracts.DependencyBindingType,
        scope: contracts.DependencyBindingScope = contracts.DependencyBindingScope.GLOBAL
    ) -> typing.NoReturn:
        contract = _auto_detect_contract(instance, contract)
        if contract:
            self._dependency_registry[contract] = contracts.DependencyBindingRecord(binding_type, instance, scope)

    def register_contractual_dependency(
        self,
        instance: contracts.DependencyBindingInstance,
        contract: contracts.DependencyBindingContract | None = None
    ) -> typing.NoReturn:
        contract = _auto_detect_contract(instance, contract)
        self.register_dependency(
            instance,
            contract,
            contracts.DependencyBindingType.CONTRACTUAL
        )

    def register_selfish_dependency(
        self,
        instance: contracts.DependencyBindingInstance,
        singleton_scope: contracts.DependencyBindingScope | None = None
    ) -> typing.NoReturn:
        binding_type = contracts.DependencyBindingType.SELFISH
        scope = contracts.DependencyBindingScope.GLOBAL

        if singleton_scope:
            binding_type = contracts.DependencyBindingType.SINGLETON
            scope = singleton_scope

        self.register_dependency(
            instance,
            instance,
            binding_type,
            scope
        )

    def register_global_singleton(
        self,
        instance: contracts.DependencyBindingInstance,
        contract: contracts.DependencyBindingContract | None = None,
    ) -> typing.NoReturn:
        self.register_dependency(
            instance,
            contract,
            contracts.DependencyBindingType.SINGLETON
        )

    def register_instance_singleton(
        self,
        instance: contracts.DependencyBindingInstance,
        contract: contracts.DependencyBindingContract | None = None,
    ) -> typing.NoReturn:
        self.register_dependency(
            instance,
            contract,
            contracts.DependencyBindingType.SINGLETON,
            contracts.DependencyBindingScope.INSTANCE
        )

    def register_callback(
        self,
        callback: typing.Callable,
        contract: contracts.DependencyBindingContract,
        event_type: contracts.EventType
    ) -> typing.NoReturn:
        self._callbacks[contract] = contracts.CallbackRecord(event_type, callback)

    def register_shutdown_callback(
        self,
        success_callback: typing.Callable,
        failure_callback: typing.Callable,
        contract: contracts.DependencyBindingContract
    ):
        self._callbacks[(contract, contracts.EventType.ON_SHUTDOWN_SUCCESS)] = success_callback
        self._callbacks[(contract, contracts.EventType.ON_SHUTDOWN_FAILURE)] = failure_callback

    def get_bindings(self) -> BindingChain:
        return BindingChain(self._dependency_registry)

    def get_provider(self) -> DependencyProvider:
        return DependencyProvider(
            bindings=self._dependency_registry,
            callbacks=self._callbacks,
            instance_registry=self._instance_registry,
        )

    async def shutdown(self, exc_type: typing.Any) -> typing.NoReturn:
        for key, callback in self._callbacks.items():
            _contract, _type = key
            _desired = contracts.EventType.ON_SHUTDOWN_FAILURE if exc_type else contracts.EventType.ON_SHUTDOWN_SUCCESS
            if _type == _desired:
                singleton_instance = self._get_from_registry(_contract)
                if exc_type:
                    await failure_shutdown_instance(callback, singleton_instance, exc_type)
                else:
                    await success_shutdown_instance(callback, singleton_instance)

    def _get_from_registry(
        self,
        contract: contracts.DependencyBindingContract
    ) -> contracts.DependencyBindingInstance:
        return self._instance_registry.get(contract)



import collections.abc
import enum
import typing
from collections import namedtuple

DependencyBindingContract = typing.Type
DependencyBindingInstance = typing.Any
DependencyBindingRecord = namedtuple('DependencyBindingRecord', ['type', 'instance', 'scope'])
DependencyBingingRegistry = collections.abc.MutableMapping[DependencyBindingContract, DependencyBindingRecord]

CallbackRegistry = collections.abc.MutableMapping[DependencyBindingContract, typing.Callable]

SingletonInstance = typing.Any
SingletonInstanceRegistry = collections.abc.MutableMapping[DependencyBindingContract, SingletonInstance]

ContextType = collections.abc.Mapping[
    tuple[DependencyBindingContract, str], typing.Any
]

DependencyBindingChainFilterClause = typing.Callable[
    [DependencyBindingContract, DependencyBindingRecord],
    bool
]

DependencyBindingChainMapClause = typing.Callable[
    [DependencyBindingContract, DependencyBindingRecord],
    typing.Any
]


class DependencyBindingType(enum.Enum):
    CONTRACTUAL = enum.auto()
    SELFISH = enum.auto()
    SINGLETON = enum.auto()


class DependencyBindingScope(enum.Enum):
    GLOBAL = enum.auto()
    INSTANCE = enum.auto()


class EventType(enum.Enum):
    ON_START = enum.auto()
    ON_SHUTDOWN_SUCCESS = enum.auto()
    ON_SHUTDOWN_FAILURE = enum.auto()


class DependencyBindingChain(typing.Protocol):
    def filter(
        self,
        clause: DependencyBindingChainFilterClause
    ) -> 'DependencyBindingChain': ...

    def map(
        self,
        key_clause: DependencyBindingChainMapClause,
        value_clause: DependencyBindingChainMapClause
    ) -> 'DependencyBindingChain': ...

    def items(self) -> collections.abc.Mapping: ...

    def keys(self) -> collections.abc.Iterable: ...

    def values(self) -> collections.abc.Iterable: ...


class DependencyProvider(typing.Protocol):
    async def resolve(
        self,
        contract: DependencyBindingContract,
        context: ContextType | None = None
    ): ...

    async def shutdown(self, exc_type: typing.Any): ...

    async def __aenter__(self): ...

    async def __aexit__(self, exc_type, exc, tb): ...


class DependencyContainer(typing.Protocol):
    def register_contractual_dependency(
        self,
        instance: DependencyBindingInstance,
        contract: DependencyBindingContract | None = None
    ) -> typing.NoReturn: ...

    def register_selfish_dependency(
        self,
        instance: DependencyBindingInstance
    ) -> typing.NoReturn: ...

    def register_global_singleton(
        self,
        instance: DependencyBindingInstance,
        contract: DependencyBindingContract | None = None,
    ) -> typing.NoReturn: ...

    def register_instance_singleton(
        self,
        instance: DependencyBindingInstance,
        contract: DependencyBindingContract | None = None,
    ) -> typing.NoReturn: ...

    def register_dependency(
        self,
        instance: DependencyBindingInstance,
        contract: DependencyBindingContract,
        binding_type: DependencyBindingType,
        scope: DependencyBindingScope = DependencyBindingScope.GLOBAL
    ) -> typing.NoReturn: ...

    def register_callback(
        self,
        callback: typing.Callable,
        contract: DependencyBindingContract,
        event_type: EventType
    ): ...

    def register_shutdown_callback(
        self,
        success_callback: typing.Callable,
        failure_callback: typing.Callable,
        contract: DependencyBindingContract,
    ): ...

    def get_bindings(self) -> DependencyBindingChain: ...

    def get_provider(self) -> DependencyProvider: ...

    async def shutdown(self, exc_type: typing.Any): ...



import inspect
import typing

from .. import contracts


async def success_shutdown_instance(
    callback_function,
    instance: contracts.SingletonInstance
):
    if not instance:
        return
    if inspect.iscoroutinefunction(callback_function):
        await callback_function(instance)
    elif callable(callback_function):
        callback_function(instance)


async def failure_shutdown_instance(
    callback_function,
    instance: contracts.SingletonInstance,
    exc_type: typing.Any
):
    if not instance:
        return
    if inspect.iscoroutinefunction(callback_function):
        await callback_function(instance, exc_type)
    elif callable(callback_function):
        callback_function(instance, exc_type)



import inspect
import typing

from .. import contracts
from .generic import success_shutdown_instance, failure_shutdown_instance


class DependencyProvider(contracts.DependencyProvider):
    def __init__(
        self,
        bindings: contracts.DependencyBingingRegistry,
        callbacks: contracts.CallbackRegistry,
        instance_registry: contracts.SingletonInstanceRegistry
    ):
        self._dependency_registry = bindings
        self._callbacks = callbacks
        self._global_instance_registry = instance_registry
        self._local_instance_registry: contracts.SingletonInstanceRegistry = {}

    async def resolve(
        self,
        contract: contracts.DependencyBindingContract,
        context: contracts.ContextType | None = None
    ):
        context = context or {}
        result = await self._resolve(contract, context=context)
        return result

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.shutdown(exc_type)

    async def shutdown(self, exc_type: typing.Any):
        for key, callback in self._callbacks.items():
            _contract, _type = key
            _desired = contracts.EventType.ON_SHUTDOWN_FAILURE if exc_type else contracts.EventType.ON_SHUTDOWN_SUCCESS
            if _type == _desired:
                singleton_instance = self._get_from_registry(
                    _contract,
                    filter_by_scope=contracts.DependencyBindingScope.INSTANCE
                )
                if exc_type:
                    await failure_shutdown_instance(callback, singleton_instance, exc_type)
                else:
                    await success_shutdown_instance(callback, singleton_instance)

    async def _resolve(
        self,
        contract: contracts.DependencyBindingContract,
        context: contracts.ContextType
    ):
        binding_record = self._dependency_registry.get(contract)
        if not binding_record:
            return

        if result := self._get_from_registry(contract):
            return result

        instance = binding_record.instance

        if inspect.iscoroutinefunction(instance):
            result = await instance()
        elif not inspect.isclass(instance):
            result = instance
        else:
            init_spec = inspect.getfullargspec(getattr(instance, '__init__'))
            params = {}
            for param_name, param_type in init_spec.annotations.items():
                if value := context.get((contract, param_name)):
                    params[param_name] = value
                else:
                    params[param_name] = await self._resolve(param_type, context=context)
            result = instance(**params)

        self._add_to_registry(contract, result)

        return result

    def _get_from_registry(
        self,
        contract: contracts.DependencyBindingContract,
        filter_by_scope: contracts.DependencyBindingScope = contracts.DependencyBindingScope.GLOBAL
    ):
        registries = [self._local_instance_registry]
        if filter_by_scope == contracts.DependencyBindingScope.INSTANCE:
            registries.append(self._global_instance_registry)

        for registry in registries:
            if contract in registry:
                return registry.get(contract)

    def _add_to_registry(self, contract: contracts.DependencyBindingContract, instance: contracts.SingletonInstance):
        binding_record = self._dependency_registry.get(contract)
        if not binding_record:
            return
        if binding_record.type != contracts.DependencyBindingType.SINGLETON:
            return

        if binding_record.scope == contracts.DependencyBindingScope.INSTANCE:
            self._local_instance_registry[contract] = instance
        else:
            self._global_instance_registry[contract] = instance
